@@ -2,14 +2,17 @@
 """
 mathicsscript is a command-line interface to Mathics.
 
-Copyright 2020-2021 The Mathics Team
+Copyright 2020-2021, 2024-2025 The Mathics3 Team
 """
+
+import os.path as osp
+
+from mathics.session import autoload_files
+
 from mathicsscript.version import __version__
 
 
 def load_default_settings_files(definitions):
-    import os.path as osp
-    from mathics.core.definitions import autoload_files
 
     root_dir = osp.realpath(osp.dirname(__file__))
 
