@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# Copyright (C) 2021 Rocky Bernstein <rb@dustyfeet.com>
+# Copyright (C) 2021-2022, 2024-2025 Rocky Bernstein <rb@dustyfeet.com>
 #   This program is free software: you can redistribute it and/or modify
 #   it under the terms of the GNU General Public License as published by
 #   the Free Software Foundation, either version 3 of the License, or
@@ -13,12 +13,29 @@
 #   You should have received a copy of the GNU General Public License
 #   along with this program.  If not, see <http://www.gnu.org/licenses/>.
 
+"""
+Keyboard input binding routines for prompt_toolkit which are
+analogous to GNU Readlines' parse_and_bind().
+"""
+
+from typing import Callable
 from prompt_toolkit.enums import EditingMode
 from prompt_toolkit.key_binding import KeyBindings
-
+from prompt_toolkit.filters import Condition
+from sys import version_info
+import contextlib
 import pathlib
 import re
 
+from mathicsscript.settings import definitions
+from mathics.session import get_settings_value
+
+
+@Condition
+def autocomplete_on():
+    return get_settings_value(definitions, "Settings`$GroupAutocomplete")
+
+
 bindings = KeyBindings()
 
 # bindings.add("escape", "p", "escape")(
@@ -26,77 +43,83 @@ bindings = KeyBindings()
 # )
 
 
-@bindings.add("{")
+@bindings.add("{", filter=autocomplete_on)
 def curly_left(event):
     b = event.cli.current_buffer
-    if event.app.group_autocomplete:
-        b.insert_text("{")
-        b.insert_text("}", move_cursor=False)
-    else:
-        b.insert_text("{")
+    b.insert_text("{")
+    b.insert_text("}", move_cursor=False)
 
 
-@bindings.add("}")
+@bindings.add("}", filter=autocomplete_on)
 def curly_right(event):
-    if event.app.group_autocomplete:
-        b = event.cli.current_buffer
-        char = b.document.current_char
-
-        if char == "}":
-            b.cursor_right()
-        else:
-            b.insert_text("}")
+    b = event.cli.current_buffer
+    char = b.document.current_char
+    if char == "}":
+        b.cursor_right()
+    else:
+        b.insert_text("}")
 
 
-@bindings.add("(")
+@bindings.add("(", filter=autocomplete_on)
 def paren_left(event):
     b = event.cli.current_buffer
-    if event.app.group_autocomplete:
-        b.insert_text("(")
-        b.insert_text(")", move_cursor=False)
+    b.insert_text("(")
+    if len(b.text) > 1 and b.text[-2] == "\\":
+        b.insert_text("\\)", move_cursor=False)
     else:
-        b.insert_text("(")
+        b.insert_text(")", move_cursor=False)
 
 
-@bindings.add(")")
+@bindings.add(")", filter=autocomplete_on)
 def paren_right(event):
     b = event.cli.current_buffer
-    if event.app.group_autocomplete:
-        char = b.document.current_char
-
-        if char == ")":
-            b.cursor_right()
-        else:
-            b.insert_text(")")
+    char = b.document.current_char
+    if char == ")":
+        b.cursor_right()
     else:
         b.insert_text(")")
 
 
-@bindings.add("[")
+@bindings.add("[", filter=autocomplete_on)
 def bracket_left(event):
     b = event.cli.current_buffer
-    if event.app.group_autocomplete:
-        b.insert_text("[")
-        b.insert_text("]", move_cursor=False)
-    else:
-        b.insert_text("[")
+    b.insert_text("[")
+    b.insert_text("]", move_cursor=False)
 
 
-@bindings.add("]")
+@bindings.add("]", filter=autocomplete_on)
 def bracket_right(event):
     b = event.cli.current_buffer
-    if event.app.group_autocomplete:
-        char = b.document.current_char
-
-        if char == "]":
-            b.cursor_right()
-        else:
-            b.insert_text("]")
+    char = b.document.current_char
+    if char == "]":
+        b.cursor_right()
     else:
         b.insert_text("]")
 
 
-# Add an additional key binding for toggling this flag.
+@bindings.add("'", filter=autocomplete_on)
+def single_quotation(event):
+    b = event.cli.current_buffer
+    char = b.document.current_char
+    if char == "'":
+        b.cursor_right()
+    else:
+        b.insert_text("'")
+        b.insert_text("'", move_cursor=False)
+
+
+@bindings.add('"', filter=autocomplete_on)
+def double_quotation(event):
+    b = event.cli.current_buffer
+    char = b.document.current_char
+    if char == '"':
+        b.cursor_right()
+    else:
+        b.insert_text('"')
+        b.insert_text('"', move_cursor=False)
+
+
+# Add an additional key binding for toggling the Edit mode flag.
 @bindings.add("f4")
 def _editor_toggle(event):
     """Toggle between Emacs and Vi mode."""
@@ -108,29 +131,36 @@ def _editor_toggle(event):
         app.editing_mode = EditingMode.VI
 
 
-# Add an additional key binding for toggling this flag.
+@bindings.add("f1")
+def _toggle_help(event):
+    """Toggles bottom bar from showing help or showing information"""
+    app = event.app
+    app.help_mode = not app.help_mode
+
+
+# Add an additional key binding for toggling the braces completion flag.
 @bindings.add("f3")
 def _group_autocomplete_toggle(event):
     """Complete braces."""
     app = event.app
-
-    if not hasattr(app, "group_autocomplete"):
-        app.group_autocomplete = False
     app.group_autocomplete = not app.group_autocomplete
 
 
-def read_inputrc(use_unicode: bool) -> None:
+def read_inputrc(read_init_file_fn: Callable, use_unicode: bool) -> None:
     """
     Read GNU Readline style inputrc
     """
-    # GNU Readling inputrc $include's paths are relative to itself,
+    # GNU Readline inputrc $include's paths are relative to itself,
     # so chdir to its directory before reading the file.
     parent_dir = pathlib.Path(__file__).parent.absolute()
-    with parent_dir:
+    path_context_fn = (
+        parent_dir if version_info < (3, 11) else contextlib.chdir(parent_dir)
+    )
+    with path_context_fn:
         inputrc = "inputrc-unicode" if use_unicode else "inputrc-no-unicode"
         try:
-            read_init_file(str(parent_dir / "data" / inputrc))
-        except:
+            read_init_file_fn(str(parent_dir / "data" / inputrc))
+        except Exception:
             pass
 
 
