@@ -3,19 +3,32 @@ Format Mathics objects
 """
 
 import math
-import networkx as nx
 import random
 from tempfile import NamedTemporaryFile
+from typing import Callable
 
+import networkx as nx
 from mathics.core.atoms import String
+from mathics.core.symbols import Symbol
 from mathics.core.systemsymbols import (
+    SymbolAborted,
     SymbolExport,
     SymbolExportString,
     SymbolFullForm,
+    SymbolGraphics,
+    SymbolGraphics3D,
+    SymbolImage,
+    SymbolMathMLForm,
+    SymbolOutputForm,
+    SymbolPlot,
     SymbolStandardForm,
+    SymbolStringForm,
+    SymbolTeXForm,
 )
 from mathics.session import get_settings_value
 
+PyMathicsGraph = Symbol("Pymathics`Graph")
+
 try:
     from matplotlib import __version__ as matplotlib_version
 except ImportError:
@@ -36,7 +49,7 @@ try:
 except ImportError:
     svg2png = None
 
-from mathicsscript.asymptote import Asy
+from mathicsscript.asymptote import Asy, write_asy_and_view
 
 have_asymptote = False
 try:
@@ -49,7 +62,14 @@ else:
 
 
 def format_output(obj, expr, format=None):
-    def eval_boxes(result, fn, obj, **options):
+    """
+    Handle unformatted output using the *specific* capabilities of mathicsscript
+
+    evaluation.py format_output() from which this was derived is similar but
+    it can't make use of a front-ends specific capabilities.
+    """
+
+    def eval_boxes(result, fn: Callable, obj, **options):
         try:
             boxes = fn(evaluation=obj, **options)
         except BoxError:
@@ -70,83 +90,101 @@ def format_output(obj, expr, format=None):
     if isinstance(format, dict):
         return dict((k, obj.format_output(expr, f)) for k, f in format.items())
 
-    from mathics.core.expression import Expression, BoxError
+    from mathics.core.expression import BoxError, Expression
 
     expr_type = expr.get_head_name()
-    if expr_type == "System`MathMLForm":
+    expr_head = expr.get_head()
+
+    if expr_head is SymbolMathMLForm:
         format = "xml"
         elements = expr.elements
         if len(elements) == 1:
             expr = elements[0]
-    elif expr_type == "System`TeXForm":
+    elif expr_head is SymbolTeXForm:
         format = "tex"
         elements = expr.elements
         if len(elements) == 1:
             expr = elements[0]
-    elif expr_type == "System`Image":
-        if get_settings_value(obj.definitions, "Settings`$UseMatplotlib") and plt:
-            temp_png = NamedTemporaryFile(
-                mode="w+b", suffix=".png", prefix="mathicsscript-"
+    elif (
+        expr_head is SymbolImage
+        and get_settings_value(obj.definitions, "Settings`$UseMatplotlib")
+        and plt
+    ):
+        temp_png = NamedTemporaryFile(
+            mode="w+b", suffix=".png", prefix="mathicsscript-"
+        )
+        try:
+            png_expr = Expression(
+                SymbolExport, String(temp_png.name), expr, String("PNG")
             )
-            try:
-                png_expr = Expression(
-                    SymbolExport, String(temp_png.name), expr, String("PNG")
-                )
-                result = png_expr.evaluate(obj)
-                plt.axes().set_axis_off()
-                img = mpimg.imread(temp_png)
-                cmap = "gray" if expr.color_space == "Grayscale" else None
-                plt.imshow(img, cmap=cmap)
-                plt.show()
-            except:  # noqa
-                pass
-            temp_png.close()
-
+            result = png_expr.evaluate(obj)
+            plt.axes().set_axis_off()
+            img = mpimg.imread(temp_png)
+            cmap = "gray" if expr.color_space == "Grayscale" else None
+            plt.imshow(img, cmap=cmap)
+            plt.show()
+        except:  # noqa
             pass
+        temp_png.close()
 
-    elif expr_type in ("System`Graphics", "System`Plot"):
-        if (
-            get_settings_value(obj.definitions, "Settings`$UseMatplotlib")
-            and plt
-            and svg2png
-        ):
-            svg_expr = Expression(SymbolExportString, expr, String("SVG"))
-            svg_str = svg_expr.evaluate(obj).to_python(string_quotes=False)
-            temp_png = NamedTemporaryFile(
-                mode="w+b", suffix=".png", prefix="mathicsscript-"
-            )
-            try:
-                svg2png(bytestring=svg_str, write_to=temp_png.name)
-                plt.axes().set_axis_off()
-                img = mpimg.imread(temp_png)
-                plt.imshow(img)
-                plt.show()
-                temp_png.close()
-            except:  # noqa
-                pass
+    elif (
+        expr_head in (SymbolGraphics, SymbolPlot)
+        and get_settings_value(obj.definitions, "Settings`$UseMatplotlib")
+        and plt
+        and svg2png
+    ):
+        svg_expr = Expression(SymbolExportString, expr, String("SVG"))
+        svg_str = svg_expr.evaluate(obj).to_python(string_quotes=False)
+        temp_png = NamedTemporaryFile(
+            mode="w+b", suffix=".png", prefix="mathicsscript-"
+        )
+        try:
+            svg2png(bytestring=svg_str, write_to=temp_png.name)
+            plt.axes().set_axis_off()
+            img = mpimg.imread(temp_png)
+            plt.imshow(img)
+            plt.show()
+            temp_png.close()
+        except:  # noqa
+            pass
         return expr_type
     elif (
-        expr_type in ("System`Graphics3D",)
+        expr_head in (SymbolGraphics, SymbolPlot, SymbolGraphics3D)
         and have_asymptote
         and get_settings_value(obj.definitions, "Settings`$UseAsymptote")
     ):
         asy_expr = Expression(SymbolExportString, expr, String("asy"))
         asy_str = asy_expr.evaluate(obj).to_python(string_quotes=False)
-        asymptote_graph.erase()
-        asymptote_graph.send(asy_str)
+
+        # Alternate older version
+        # asymptote_graph.erase()
+        # asymptote_graph.send(asy_str)
+
+        write_asy_and_view(asy_str)
         return expr_type
 
+    if expr is SymbolAborted:
+        obj.out = ["$Aborted"]
+        obj.last_eval = SymbolAborted
+        return "$Aborted"
     if format == "text":
-        result = expr.format(obj, "System`OutputForm")
+        if isinstance(expr, String):
+            return expr.value
+        result = expr.format(obj, SymbolOutputForm)
     elif format == "xml":
-        result = Expression(SymbolStandardForm, expr).format(obj, "System`MathMLForm")
+        result = Expression(SymbolStandardForm, expr).format(obj, SymbolMathMLForm)
     elif format == "tex":
-        result = Expression(SymbolStandardForm, expr).format(obj, "System`TeXForm")
+        result = Expression(SymbolStandardForm, expr).format(obj, SymbolTeXForm)
+    elif format == "error":
+        # We may do fancier things in the future.
+        if isinstance(expr, String):
+            return expr.value
+        result = expr.format(obj, SymbolOutputForm)
     elif format == "unformatted":
-        if str(expr) == "-Graph-" and hasattr(expr, "G"):
+        if expr_head is PyMathicsGraph and hasattr(expr, "G"):
             return format_graph(expr.G)
         else:
-            result = expr.format(obj, "System`OutputForm")
+            result = expr.format(obj, SymbolOutputForm)
     else:
         raise ValueError
 
@@ -168,7 +206,6 @@ cached_pair = None
 def hierarchy_pos(
     G, root=None, width=1.0, vert_gap=0.2, vert_loc=0, leaf_vs_root_factor=0.5
 ):
-
     """Position nodes in tree layout. The root is at the top.
 
     Based on Joel's answer at https://stackoverflow.com/a/29597209/2966723,
@@ -245,7 +282,7 @@ def hierarchy_pos(
         return cached_pair
 
     # These get swapped if tree edge directions point to the root.
-    decendants = nx.descendants
+    descendants = nx.descendants
     out_degree = G.out_degree if hasattr(G, "out_degree") else G.degree
     neighbors = G.neighbors
 
@@ -257,7 +294,7 @@ def hierarchy_pos(
                 # The case where we have a one or two node graph is ambiguous.
                 root = list(nx.topological_sort(G))[-1]
                 # Swap motion functions
-                decendants = nx.ancestors
+                descendants = nx.ancestors
                 out_degree = G.in_degree
                 neighbors = G.predecessors
             else:
@@ -332,7 +369,9 @@ def hierarchy_pos(
 
     xcenter = width / 2.0
     if isinstance(G, nx.DiGraph):
-        leafcount = len([node for node in decendants(G, root) if out_degree(node) == 0])
+        leafcount = len(
+            [node for node in descendants(G, root) if out_degree(node) == 0]
+        )
     elif isinstance(G, nx.Graph):
         leafcount = len(
             [
