@@ -1,6 +1,5 @@
 # -*- coding: utf-8 -*-
 
-import pytest
 from mathics.core.definitions import Definitions
 from mathicsscript.termshell_gnu import TerminalShellGNUReadline
 
@@ -16,7 +15,6 @@ def test_completion_gnu():
     definitions = Definitions(add_builtin=True, extension_modules=[])
     term = TerminalShellGNUReadline(
         definitions=definitions,
-        style=None,
         want_readline=True,
         want_completion=True,
         use_unicode=False,
@@ -34,4 +32,4 @@ def test_completion_gnu():
 
 
 if __name__ == "__main__":
-    test_completion()
+    test_completion_gnu()
